@@ -8,262 +8,197 @@
 
 #include "../inc/services.h"
 
-Channel *chanlist = NULL;
+Channel *chanlist[1024];
+#define HASH(chan)      ((chan)[1] ? ((chan)[1]&31)<<5 | ((chan)[2]&31) : 0)
 
+static int schans = 0;
 #ifdef SJOIN
 static int smembers = 1;
 static int ssize = 0;
-static int schans = 0;
 
 static struct sjoin_list {
    char *nick;
 } *slist = NULL;
 static void extract_nicks();
-void get_sjoin_list(char *list, int newchan);
-static void add_suser(char *nick, char *channel);
-static void scheck_ops(char *chan, char **checklist, int count);
-static void update_last_used(const char *chan);
+static void add_suser(User *user, Channel *chan);
 #endif
 
-#ifdef SJOIN
+void get_sjoin_list(char *list, int newchan);
+static void chan_delete(Channel *c);
+static void chan_adduser2(User *user, Channel *c);
+static Channel *chan_create(const char *chan);
+
 /*************************************************************************/
+#ifdef SJOIN
+
 void load_sjoin_mem()
 {
        slist = smalloc(sizeof(*slist) * 1024);
        ssize = 1024;
 }
 /*************************************************************************/
+/*
+SJOIN:
+	av[0]	Time
+	av[1]	chan
+	av[2]	mode or +
+	av[3]	none
+	av[4]	cmodenick
+SJ3
+	av[2]	Mode || User
+	av[3]	user || Null
+*/
+
 void do_sjoin(char *source, int ac, char **av)
 {
    Channel *chan;
-#if (BAH_V >= 150) || defined(RAGEIRCD)
-   int sf = 1;
-#else
    int sf = 0;
-#endif
-   char *schan = av[2-sf];
-   char *s, *who, *tmpwho;
-   char *mv[3], *checklist[1024];
-   int args = 0, start = 4;
-   int newchan = 0, bogus=0, check=0, inmean = 0, x = 0;
+   char *s, *nick, *t;
+   int start = 4;
+   int newchan = 0, x = 0;
+   int modecnt;
    User *user;
+
+#ifdef SJOIN
+	sf = 0;
+#endif
+#ifdef SJ3
+	sf = 1;
+#endif
+#ifdef SSJ3
+	sf = 1;
+#endif
+
    chan = findchan(av[2-sf]);
+
    if (!chan) {
       if (debug)
            log("Creating new channel %s via sjoin", av[2-sf]);
-
-      chan = scalloc(sizeof(Channel), 1);
-      chan->next = chanlist;
-      if (chanlist)
-          chanlist->prev = chan;
-      chanlist = chan;
-      strscpy(chan->name, av[2-sf], sizeof(chan->name));
-      chan->creation_time = CTime;
-      chan->ngames = 1;
-      chan->wgames = 1;
-      newchan = 1;
+	chan = chan_create(av[2-sf]);
+	newchan = 1;
    }
 
-
-#if BAH_V >= 150
-   if (!strchr(source, '.')) {
-        if (!(user = finduser(source))) {
-             bogus=0; smembers=0;
-             goto bogus_join;
-        }
-        chan_adduser(user, schan);
-        add_suser(source, chan->name);
-        mv[0] = sstrdup(schan); mv[1] = "+o"; mv[2] = sstrdup(source);
-        do_cmode(source, 3, mv);
-        free(mv[0]); free(mv[2]);
-        return;
+   if (*av[3-sf] != '+' && *av[3-sf] != '-') {
+#ifdef SJ3
+      start = start - 1;
+#endif
+      goto skipmodes;
    }
-#endif
 
-#ifdef RAGEIRCD
-	sf = 0;
-#endif
-
-   if (stricmp(av[3-sf], "0") == 0)
+   if (!stricmp(av[3-sf], "+"))
       goto skipmodes;
 
    s = av[3-sf];
+
    strscpy(chan->modes, changemode(s, chan->modes), sizeof(chan->modes));
+
    while (*s) {
         switch(*s++) {
             case 'l':
                 if (chan->limit > 0)
                    chan->limit = 0;
-                chan->limit = atoi(av[4-sf]);
-		inmean++;
+                chan->limit = atoi(av[start-sf]);
+		start++;
                 break;
             case 'k':
                 if (chan->key) {
                    free(chan->key);
-                   chan->key = NULL;
                 }
-		inmean++;
-                chan->key = sstrdup(av[4-sf+inmean]);
-                args++;
+                chan->key = sstrdup(av[start-sf]);
+		start++;
                 break;
+#ifdef CFLOOD
+            case 'f':
+                if (chan->flood) {
+                   free(chan->flood);
+                }
+                chan->flood = sstrdup(av[start-sf]);
+		start++;
+                break;
+#endif
+#ifdef CLINK
+            case 'L':
+                if (chan->link) {
+                   free(chan->link);
+                }
+                chan->link = sstrdup(av[start-sf]);
+		start++;
+                break;
+#endif
         }
    }
 
-   strscpy(chan->modes, changemode("-lk", chan->modes), sizeof(chan->modes));
-   check_modes(chan->name, source);
-
-   if (stricmp(av[3-sf], "+") != 0) {
-
-       if (chan->key && (chan->limit > 0))
-           start += 2;
-       else if ((!chan->key && (chan->limit > 0)) ||
-                 (chan->key && (chan->limit == 0)))
-           start++;
-   }
-
 skipmodes:
 
-   mv[0] = sstrdup(chan->name);
+   check_modes(chan->name, source);
+
    get_sjoin_list(av[start-sf], newchan);
 
    for (x = 0; x < smembers; x++) {
+ 	 char modebuf2[7] = { 0 };
+	 t = slist[x].nick;
+	 while (*(nick = t)) {
+	        t = nick + strcspn (nick, " ");
+	        if (*t)
+	            *t++ = 0;
 
-         if (strchr(slist[x].nick, '@') && strchr(slist[x].nick, '+')) {
-       	     if (!(user = finduser(strtok(slist[x].nick, "@+")))) {
-                bogus++; continue;
-             }
-             mv[2] = sstrdup(user->nick);
-             add_suser(mv[2], mv[0]);
-             if (check_valid_op(user, mv[0], 1)) {
-                mv[1] = "+o";
-                do_cmode(source, 3, mv);
-             } else {
-                 send_cmd(s_ChanServ, "MODE %s -o %s", mv[0], mv[2]);
-                 notice(s_ChanServ, mv[2], toknmsg[25], chan->name, s_ChanServ);
-             }
+	        modecnt = 0;
+	        modebuf2[0] = '+';
 
-             mv[1] = "+v";
-             do_cmode(source, 3, mv);
+		while (*nick == '@' || *nick == '+' || *nick == '%' || *nick == '*' || *nick == '~')
+		{
+			switch (*nick)
+			{
+			case '@':
+		                modecnt++;
+                	        modebuf2[modecnt] = 'o';
+		                break;
+                	case '+':
+	                        modecnt++;
+             	                modebuf2[modecnt] = 'v';
+	                        break;
+                	case '%':
+	                        modecnt++;
+             	                modebuf2[modecnt] = 'h';
+	                        break;
+                	case '*':
+	                        modecnt++;
+             	                modebuf2[modecnt] = 'q';
+	                        break;
+                	case '~':
+	                        modecnt++;
+             	                modebuf2[modecnt] = 'a';
+	                        break;
+		        }
+	        	nick++;
+	     	} /* mode */
+	/* if match in ban or ban exception */
+	     if (*nick != '&' && *nick !='"') {
+		     user = finduser(nick);
+		     if (user) {
+			if (!check_kick(user, chan->name, chan) && !check_invite(user, chan->name)) {
+	        	     add_suser(user, chan);
+			     check_welcome(user, chan->name);
+			     checkmass_shouldop(user, chan, modebuf2);
+			}
+		     } else {
+			     log("Try sjoin nonexist user %s", nick);
+		     } /* user */
+	     } else {
+		     if (*nick == '&') {
+			char *avs[3];
+			avs[0] =  chan->name;
+			avs[1] = "+b";
+			*nick = *nick++;
+			avs[2] = sstrdup(nick);
+			do_cmode(s_ChanServ, 3, avs);
+			free(avs[2]);
+		     }	
+	     } /* Ban and ban exception */
 
-	 } else if (strchr(slist[x].nick, '@')) {
-             if (!(user = finduser(strtok(slist[x].nick, "@")))) {
-                bogus++; continue;
-             }
-             mv[2] = sstrdup(user->nick);
-             add_suser(mv[2], mv[0]);
-             if (check_valid_op(user, mv[0], 1)) {
-	         mv[1] = sstrdup("+o");
-       	         do_cmode(source, 3, mv);
-             } else {
-                 send_cmd(s_ChanServ, "MODE %s -o %s", mv[0], mv[2]);
-                 notice(s_ChanServ, mv[2], toknmsg[25], chan->name, s_ChanServ);
-             }
-#ifdef DEFHALFOP
-	 } else if (strchr(slist[x].nick, '%')) {
-             if (!(user = finduser(strtok(slist[x].nick, "%")))) {
-                bogus++; continue;
-             }
-             mv[2] = sstrdup(user->nick);
-             mv[1] = sstrdup("+h");
-             add_suser(mv[2], mv[0]);
-             do_cmode(source, 3, mv);
-#endif
-#ifdef DEFPROTECT
-	 } else if (strchr(slist[x].nick, '^')) {
-             if (!(user = finduser(strtok(slist[x].nick, "^")))) {
-                bogus++; continue;
-             }
-             mv[2] = sstrdup(user->nick);
-             mv[1] = sstrdup("+a");
-             add_suser(mv[2], mv[0]);
-             do_cmode(source, 3, mv);
-#endif
-#ifdef UNREAL
-	} else if (strchr(slist[x].nick, '*')) {
-             if (!(user = finduser(strtok(slist[x].nick, "*")))) {
-                bogus++; continue;
-             }
-             mv[2] = sstrdup(user->nick);
-             mv[1] = sstrdup("+q");
-             add_suser(mv[2], mv[0]);
-             do_cmode(source, 3, mv);
-
-#endif
-
-	 } else if (strchr(slist[x].nick, '+')) {
-             if (!(user = finduser(strtok(slist[x].nick, "+")))) {
-                bogus++; continue;
-             }
-             mv[2] = sstrdup(user->nick);
-             mv[1] = sstrdup("+v");
-             add_suser(mv[2], mv[0]);
-             do_cmode(source, 3, mv);
-
-         } else {
-             if (!(user = finduser(slist[x].nick))) {
-                bogus++; continue;
-             }
-             mv[2] = sstrdup(slist[x].nick);
-             add_suser(mv[2], mv[0]);
-	}
-
-         if ((bogus == smembers) && newchan)
-             goto bogus_join;
-
-      user = finduser(mv[2]);
-
-      if (!user)
-	return;
-
-      if (!is_chanop(mv[2], schan))
-           chan_adduser(user, schan);
-
-      free(slist[x].nick); slist[x].nick = sstrdup(mv[2]);
-      free(mv[2]);
-
-      check_welcome(user, chan->name);
-
-/* do_join removed - as I said, I got around to improving it :P */
-
-      if (newchan && (x < 1)) { /* x to prevent repeated topics */
-            restore_topic(chan->name);
-#if defined(BAHAMUT) || defined(RAGEIRCD)
-	    if (!chan->havebot)
-		    botjoin(chan->name);
-	    do_cs_join(chan->name);
-#endif
-
-      }
-
-      mv[1] = NULL;
+	 } /* each nick */
    }
-   if (checklist[0]) {
-      scheck_ops(mv[0], checklist, check);
-      checklist[0] = 0;
-   }
-   free(mv[0]);
-
-bogus_join:
-
-   if ((bogus == smembers) && newchan) {
-       if (chan->next)
-            chan->next->prev = chan->prev;
-       if (chan->prev)
-            chan->prev->next = chan->next;
-       else
-            chanlist = chan->next;
-       free(chan); free(mv[2]);
-   }
-
-   for (x = 0; x < smembers; x++) {
-       user = finduser(slist[x].nick);
-       if (user) {
-           check_kick(user, chan->name);
-           check_invite(user, chan->name);
-       }
-           free(slist[x].nick);
-   }
-
+   for (x = 0; x < smembers; x++) free(slist[x].nick);
 }
 
 /*************************************************************************/
@@ -288,7 +223,6 @@ void extract_nicks()
    while ((tmp = strtok(NULL, " ")) != NULL)
    {
        slist[i++].nick = sstrdup(tmp);
-
    }
    smembers = i;
   
@@ -301,39 +235,48 @@ void extract_nicks()
 
 void get_channel_stats(long *nrec, long *memuse)
 {
-    long count = 0, mem = 0;
-    Channel *chan;
-    struct c_userlist *cu;
-    int i;
+	long count = 0, mem = 0;
+	Channel *chan;
+	struct c_userlist *cu;
+	BanData *bd;
+	int i, j;
+
+	for (j = 0; j < 1024; j++) {
+	for (chan = chanlist[i]; chan; chan = chan->next) {
+		count++;
+		mem += sizeof(*chan);
+		if (chan->topic)
+		    mem += strlen(chan->topic)+1;
+		if (chan->key)
+		    mem += strlen(chan->key)+1;
+		if (chan->flood)
+		    mem += strlen(chan->flood)+1;
+		if (chan->link)
+		    mem += strlen(chan->link)+1;
+		if (chan->wgame)
+		    mem += strlen(chan->wgame)+1;
+		mem += sizeof(char *) * chan->bansize;
+		for (i = 0; i < chan->bancount; i++) {
+		    if (chan->bans[i])
+			mem += strlen(chan->bans[i])+1;
+		}
+		for (cu = chan->users; cu; cu = cu->next) {
+        	    if (cu->ud) {
+                	mem += sizeof(*cu->ud);
+	               if (cu->ud->lastline) mem += strlen(cu->ud->lastline)+1;
+        	    }
+	 		mem += sizeof(*cu);
+		}
+        	for (bd = chan->bd; bd; bd = bd->next)
+	        {
+        	        if (bd->mask) mem += strlen(bd->mask)+1;
+	                mem+= sizeof(*bd);
+        	}
 
-    for (chan = chanlist; chan; chan = chan->next) {
-	count++;
-	mem += sizeof(*chan);
-	if (chan->topic)
-	    mem += strlen(chan->topic)+1;
-	if (chan->key)
-	    mem += strlen(chan->key)+1;
-	if (chan->flood)
-	    mem += strlen(chan->flood)+1;
-	if (chan->link)
-	    mem += strlen(chan->link)+1;
-	if (chan->wgame)
-	    mem += strlen(chan->wgame)+1;
-	mem += sizeof(char *) * chan->bansize;
-	for (i = 0; i < chan->bancount; i++) {
-	    if (chan->bans[i])
-		mem += strlen(chan->bans[i])+1;
 	}
-	for (cu = chan->users; cu; cu = cu->next) {
-            if (cu->ud) {
-                mem += sizeof(*cu->ud);
-               if (cu->ud->lastline) mem += strlen(cu->ud->lastline)+1;
-            }
-	 	mem += sizeof(*cu);
 	}
-    }
-    *nrec = count;
-    *memuse = mem;
+	    *nrec = count;
+	    *memuse = mem;
 }
 
 /*************************************************************************/
@@ -343,7 +286,8 @@ void get_channel_stats(long *nrec, long *memuse)
 
 Channel *findchan(const char *chan)
 {
-    Channel *c = chanlist;
+    Channel *c;
+    c = chanlist[HASH(chan)];
 
     while (c) {
 	if (stricmp(c->name, chan) == 0)
@@ -353,6 +297,17 @@ Channel *findchan(const char *chan)
     return NULL;
 }
 /*************************************************************************/
+int get_chan_status(User *u, Channel *c) {
+        struct u_chanlist *uc;
+
+        if (!u || !c) return 0;
+
+        for (uc = u->chans; uc; uc = uc->next)
+                if (uc->chan == c)
+                        return (uc->status);
+        return 0;
+}
+
 int has_chan_status(User *u, Channel *c, int status)
 {
         struct u_chanlist *uc;
@@ -393,89 +348,49 @@ void set_chan_status(User *u, Channel *c, int status)
         }
 }
 
+/*************************************************************************/
+static Channel *chan_create(const char *chan) {
+	Channel *c, **list;
+
+	if (debug)
+	    log("debug: Creating channel %s", chan);
+
+	/* Allocate pre-cleared memory */
+	c = scalloc(sizeof(Channel), 1);
+	strscpy(c->name, chan, sizeof(c->name));
+	list = &chanlist[HASH(c->name)];
+	c->next = *list;
+        if (*list)
+            (*list)->prev = c;
+        *list = c;
+	c->creation_time = CTime;
+	/* This will restore locked modes and saved topic */
+	restore_topic(chan, c);
+	check_modes(chan, s_ChanServ);
+	botjoin(c->name);
+        do_cs_join(chan);
+	return c;
+}
 /*************************************************************************/
 /* Add/remove a user to/from a channel, creating or deleting the channel as
  * necessary. */
 
-void chan_adduser(User *user, const char *chan)
-{
+void chan_adduser(User *user, const char *chan) {
     Channel *c = findchan(chan);
-    int newchan = !c;
-    struct c_userlist *u;
-
-#ifndef BAHSJOIN
-    if (newchan) {
-	if (debug)
-	    log("debug: Creating channel %s", chan);
-	/* Allocate pre-cleared memory */
-
-	c = scalloc(sizeof(Channel), 1);
-	c->next = chanlist;
-	if (chanlist)
-	    chanlist->prev = c;
-	chanlist = c;
-	strscpy(c->name, chan, sizeof(c->name));
-	c->creation_time = CTime;
-	/* This will restore locked modes and saved topic */
-        check_invite(user, chan);
-	check_modes(chan,user->nick);
-	restore_topic(chan);
-	botjoin(c->name);
-        do_cs_join(chan);
+    if (!c) {
+	c = chan_create (chan);
     }
-#endif
-
-#ifndef BAHSJOIN
-    u = scalloc(sizeof(struct c_userlist),1);
-    u->next = c->users;
-    u->prev = NULL;
-    if (c->users)
-	c->users->prev = u;
-    c->users = u;
-    u->user = user;
-    u->ud = scalloc(sizeof(UserData), 1);
-    u->ud->last_use = CTime;
-    c->usercount++;
-#endif
-
-#ifdef UNREAL
-        if (check_should_founderize(user, chan)) {
-        	set_chan_status(user, c, CSTATUS_OWNER);
-	} else
-#endif
-#ifdef DEFPROTECT
-	if (check_should_protect(user, chan)) {
-	        set_chan_status(user, c, CSTATUS_PROTECT);
-        }
-#endif
-	if (!check_opguard(user, chan, c->usercount))
-     	;
-	else if (check_should_op(user, chan)) {
-        	set_chan_status(user, c, CSTATUS_OP);
-	} 
-#ifdef DEFHALFOP
-	if (check_should_halfop(user, chan)) {
-	        set_chan_status(user, c, CSTATUS_HALFOP);
-	} else 
-#endif
-	if (check_should_voice(user, chan)) {
-	        set_chan_status(user, c, CSTATUS_VOICE);
-	}
-
+   chan_adduser2(user, c);
 }
 
+
 void chan_deluser(User *user, Channel *c)
 {
     struct c_userlist *u;
-    struct c_userlist *u2;
 
     if (!user || !c)
        return;
 
-    u2 = c->users;
-    if (!u2)
-      return;
-
     for (u = c->users; u && u->user != user; u = u->next)
         ;
     if (!u)
@@ -483,32 +398,45 @@ void chan_deluser(User *user, Channel *c)
 
     if (u->ud)
     {
-        if (u->ud->lastline) {
-                free(u->ud->lastline);
-	}
+        if (u->ud->lastline)  free(u->ud->lastline);
         free(u->ud);
     }
 
     if (u->next)
 	u->next->prev = u->prev;
+
     if (u->prev)
 	u->prev->next = u->next;
     else
 	c->users = u->next;
     free(u);
 
-    if (c->usercount)
-	    c->usercount--;
+    c->usercount--;
 
     if (!c->users) {
+	chan_delete(c);
+    }
+}
+
+/*************************************************************************/
+static void chan_delete(Channel *c) {
+	BanData *bd, *next;
 #ifdef SJOIN
         --schans;
 #endif
 	do_cs_part(c->name);
-	botpart(c->name);
+	botpart(c);
+
 	if (debug)
 	    log("debug: Deleting channel %s", c->name);
 
+        for (bd = c->bd; bd; bd = next) {
+            if (bd->mask)
+                free(bd->mask);
+            next = bd->next;
+            free(bd);
+        }
+
 	if (c->topic)
 	    free(c->topic);
 	if (c->key)
@@ -532,13 +460,10 @@ void chan_deluser(User *user, Channel *c)
 	if (c->prev)
 	    c->prev->next = c->next;
 	else
-	    chanlist = c->next;
+	    chanlist[HASH(c->name)] = c->next;
 	free(c);
-    }
 }
-
 /*************************************************************************/
-
 /* Handle a channel MODE command. */
 
 void do_cmode(const char *source, int ac, char **av)
@@ -599,7 +524,7 @@ void do_cmode(const char *source, int ac, char **av)
 
 	    break;
 
-#ifdef UNREAL
+#ifdef CLINK
         case 'L':
             if (--ac < 0) {
                 log("channel: MODE %s %s: missing parameter for %cL",
@@ -618,7 +543,7 @@ void do_cmode(const char *source, int ac, char **av)
 
             break;
 #endif
-#if defined(UNREAL)  || defined(ULTIMATE)
+#ifdef CFLOOD
         case 'f':
             if (--ac < 0) {
                 log("channel: MODE %s %s: missing parameter for %cf",
@@ -650,6 +575,7 @@ void do_cmode(const char *source, int ac, char **av)
 					sizeof(char *) * chan->bansize);
 		}
 		chan->bans[chan->bancount++] = sstrdup(*av++);
+		checktban(chan);
 	    } else {
 		char **s1 = chan->bans;
 		int i = 0;
@@ -661,11 +587,14 @@ void do_cmode(const char *source, int ac, char **av)
 		    if (i < chan->bancount)
 			bcopy(s1+1, s1, sizeof(char *) * (chan->bancount-i));
 		}
+		deltban(chan, *av);
 		++av;
+		checktban(chan);
 	    }
 
 	    break;
 	case 'o':
+	    checktban(chan);
 	    if (--ac < 0) {
 		log("channel: MODE %s %s: missing parameter for %co",
 					chan->name, modestr, add ? '+' : '-');
@@ -682,8 +611,7 @@ void do_cmode(const char *source, int ac, char **av)
 
 		if (debug)
 		    log("debug: Setting +o on %s for %s", chan->name, nick);
-
-		set_chan_status(user, chan, CSTATUS_OP);
+		    set_chan_status(user, chan, CSTATUS_OP);
 	    } else {
 		user = finduser(nick);
 		if (!user)
@@ -699,6 +627,7 @@ void do_cmode(const char *source, int ac, char **av)
 					chan->name, modestr, add ? '+' : '-');
 		break;
 	    }
+	    checktban(chan);
 	    nick = *av++;
 	    if (add) {
 		user = finduser(nick);
@@ -719,6 +648,7 @@ void do_cmode(const char *source, int ac, char **av)
 	    break;
 #endif
 	case 'v':
+	    checktban(chan);
 	    if (--ac < 0) {
 		log("channel: MODE %s %s: missing parameter for %cv",
 					chan->name, modestr, add ? '+' : '-');
@@ -742,7 +672,7 @@ void do_cmode(const char *source, int ac, char **av)
 		remove_chan_status(user, chan, CSTATUS_VOICE);
 	    }
 	    break;
-#ifdef UNREAL
+#ifdef CFOUNDER
 	case 'q':
 	    if (--ac < 0) {
 		log("channel: MODE %s %s: missing parameter for %cq",
@@ -775,9 +705,9 @@ void do_cmode(const char *source, int ac, char **av)
 	    break;
 #endif
 #ifdef DEFPROTECT
-#if defined(UNREAL) || defined(RAGEIRCD) || defined(ULTIMATE)
+#ifdef PROTECTA
 	case 'a':
-#elif defined(LIQUID)
+#elif defined(PROTECTU)
 	case 'u':
 #endif
 	    if (--ac < 0) {
@@ -797,10 +727,10 @@ void do_cmode(const char *source, int ac, char **av)
 		    log("debug: Setting +a on %s for %s", chan->name, nick);
 
                 if (!check_valid_op(user, chan->name, !!strchr(source, '.'))) {
-#ifdef UNREAL
+#ifdef PROTECTA
                        send_cmd(s_ChanServ, "MODE %s -a %s", chan->name, nick);
 #endif
-#if defined(LIQUID) || defined(RAGEIRCD)
+#ifdef PROTECTU
                        send_cmd(s_ChanServ, "MODE %s -u %s", chan->name, nick);
 #endif
                        break;
@@ -845,24 +775,26 @@ void do_topic(const char *source, int ac, char **av)
     if (!c)
 	return;
     
+    if (!strchr(source, '.')) {
+        if (check_topiclock(source, av[0]))
+		return;
+    } else {
+	return;
+    }
+
     strscpy(c->topic_setter, av[1], sizeof(c->topic_setter));
 
     c->topic_time = atol(av[2]);
 
     if (ac > 3 && *av[3]) {
-	if (c->topic)
+	if (c->topic) {
 		free(c->topic);
+		c->topic = NULL;
+	}
 	c->topic = sstrdup(av[3]);
 
-    }
-    if (!strchr(source, '.')) {
-        if (check_topiclock(source, av[0])) {
-            return;
-	}
-	record_topic(av[0]);
-    } else {
-	return;
-    }
+     }
+     record_topic(av[0], c);
 }
 
 /*************************************************************************/
@@ -878,7 +810,7 @@ int only_one_user(const char *chan)
 {
     Channel *c;
 
-    for (c = chanlist; c; c = c->next) {
+    for (c = chanlist[HASH(chan)]; c; c = c->next) {
 	if (stricmp(c->name, chan) == 0)
 	    return (c->users && !c->users->next) ? 1 : 0;
     }
@@ -889,28 +821,19 @@ int only_one_user(const char *chan)
 
 int isin_chan(char *nick, const char *chan)
 {
-    Channel *c;
+    Channel *c = findchan(chan);
     struct c_userlist *cu;
-
-    for (c = chanlist; c; c = c->next) {
-        if (stricmp(c->name, chan) == 0) {
-	    for (cu = c->users; cu; cu = c->users->next) {
-		if (stricmp(cu->user->nick, nick) == 0)
-		    return 1;
-	    }
-	    return 0;
-	}
+    if (!c) return 0;
+    for (cu = c->users; cu; cu = c->users->next) {
+	if (stricmp(cu->user->nick, nick) == 0)
+	    return 1;
     }
     return 0;
 }
 /*************************************************************************/
 int channel_form()
 {
-    Channel *c;
-    int i = 0;
-    for (c = chanlist; c; c = c->next)
-	i++;
-    return i;
+    return schans;
 }
 /*************************************************************************/
 void change_cmode (const char *who, const char *chan, const char *mode, const char *pram)
@@ -940,35 +863,32 @@ void kick_user (const char *who, const char *chan, const char *nick, const char
 }
 /***************************************************/
 #ifdef SJOIN
-static void add_suser(char *nick, char *channel)
+static void add_suser(User *user, Channel *chan)
 {
-   Channel *chan = findchan(channel);
-   struct c_userlist *u;
-   struct u_chanlist *c;
-   User *user = finduser(nick);
-
-      u = smalloc(sizeof(struct c_userlist));
-      u->next = chan->users;
-      u->prev = NULL;
-      if (chan->users)
-         chan->users->prev = u;
-      chan->users = u;
-      u->user = user;
-
-      c = smalloc(sizeof(*c));
-      c->next = user->chans;
-      c->prev = NULL;
-      if (user->chans)
-          user->chans->prev = c;
-      user->chans = c;
-      c->chan = findchan(chan->name);
+	struct u_chanlist *c;
+	c = scalloc(sizeof(*c),1);
+	c->next = user->chans;
+	if (user->chans)
+		user->chans->prev = c;
+	user->chans = c;
+	c->chan = chan;
+	chan_adduser2(user, chan);
 }
 #endif
+static void chan_adduser2(User *user, Channel *c)
+{
+    struct c_userlist *u;
+    u = scalloc(sizeof(struct c_userlist),1);
+    u->next = c->users;
+    if (c->users)
+	c->users->prev = u;
+    c->users = u;
+    u->user = user;
+    c->usercount++;
+}
 /***************************************************************/
-char *filter_modes (char *chan, char *on, char *off)
+char *filter_modes (ChannelInfo *ci, Channel *c, char *on, char *off)
 {
-    Channel *c = findchan (chan);
-    ChannelInfo *ci = cs_findchan (chan);
     char *current = c->modes;
     char new_on[64];
     char new_off[64];
@@ -994,13 +914,11 @@ char *filter_modes (char *chan, char *on, char *off)
     if (strchr (off, 'k') && strchr (current, 'k') && !strchr (on, 'k'))
         strcat (new_off, "k");
 
-
     if (strchr (on, 'l') && !strchr (current, 'l') && !strchr (off, 'l'))
         strcat (new_on, "l");
     if (strchr (off, 'l') && strchr (current, 'l') && !strchr (on, 'l'))
         strcat (new_off, "l");
 
-
     if (strchr (on, 'm') && !strchr (current, 'm') && !strchr (off, 'm'))
         strcat (new_on, "m");
     if (strchr (off, 'm') && strchr (current, 'm') && !strchr (on, 'm'))
@@ -1075,72 +993,116 @@ char *filter_modes (char *chan, char *on, char *off)
 }
 
 /***************************************************/
-#ifdef SJOIN
-static void scheck_ops(char *chan, char **checklist, int count)
-{
-  ChannelInfo *ci = cs_findchan(chan);
-  Channel *c = findchan(chan);
-  struct c_userlist *u;
-  char buf[1024], buf2[128];
-  User *user;
-  int i, modes=0, acc=0;
-
-  *buf=0; *buf2=0;
-
-  if (!ci)
-    return;
-
-  for (i = 0; i < count; i++) {
-    user = finduser(checklist[i]);
-    if (!user)
-        continue;
-
-    acc = get_access(user, ci);
-
-    if (acc == 3) {
-        if (*buf)
-           strcat(buf, " ");
-        else
-           strcat(buf2, "+");
-        strcat(buf, user->nick);
-        strcat(buf2, "v");
-        modes++;
-	if (c)
-		set_chan_status(user, c, CSTATUS_VOICE);
+void deltban(Channel *c, char *mask) {
+    BanData *bd, *next;
+    for (bd = c->bd; bd; bd = next) {
+	if (!stricmp(bd->mask, mask)) {
+	        log("Clear tban in %s %s", c->name, mask);
+		if (bd->next)
+			bd->next->prev = bd->prev;
+		if (bd->prev)
+			bd->prev->next = bd->next;
+		else
+			c->bd = bd->next;
+		if (bd->mask) {
+			free(bd->mask);
+		}
+		next = bd->next;
+                free(bd);
+		continue;
+	} /* If */
+	next = bd->next;
     }
-    if (acc > 3) {
-        update_last_used(chan);
-        if (*buf)
-           strcat(buf, " ");
-        else
-           strcat(buf2, "+");
-
-
-        strcat(buf, user->nick);
-        strcat(buf2, "o");
-        modes++;
-	if (c)
-		set_chan_status(user, c, CSTATUS_OP);
-    }
-    if (modes > 5) {
-       send_cmd(s_ChanServ, "MODE %s %s %s", chan, buf2, buf);
-       modes = 0;
-       *buf = 0; *buf2 = 0;
-    }
-  }
-  if (*buf)
-     send_cmd(s_ChanServ, "MODE %s %s %s", chan, buf2, buf);
 }
-#endif
-/*************************************************/
-#ifdef SJOIN
-static void update_last_used(const char *chan)
-{
-   ChannelInfo *ci = cs_findchan(chan);
-
-   if (!ci)
-      return;
-   else
-      ci->last_used = CTime;
+void delttban(char *who, Channel *c, char *mask) {
+    BanData *bd, *next;
+    for (bd = c->bd; bd; bd = next) {
+	if (!match_wild_nocase(bd->mask, mask)) {
+		change_cmode(who, c->name, "-b", bd->mask);
+		if (bd->next)
+			bd->next->prev = bd->prev;
+		if (bd->prev)
+			bd->prev->next = bd->next;
+		else
+			c->bd = bd->next;
+		if (bd->mask) {
+			free(bd->mask);
+		}
+		next = bd->next;
+                free(bd);
+		continue;
+	} /* If */
+	next = bd->next;
+    }
+}
+void deltuban(char *who, Channel *c, User *u) {
+    BanData *bd, *next;
+    for (bd = c->bd; bd; bd = next) {
+	if (match_usermask(bd->mask, u)) {
+		change_cmode(who, c->name, "-b", bd->mask);
+		if (bd->next)
+			bd->next->prev = bd->prev;
+		if (bd->prev)
+			bd->prev->next = bd->next;
+		else
+			c->bd = bd->next;
+		if (bd->mask) {
+			free(bd->mask);
+		}
+		next = bd->next;
+                free(bd);
+		continue;
+	} /* If */
+	next = bd->next;
+    }
+}
+void checktban(Channel *c) {
+    BanData *bd, *next;
+    time_t now = time(NULL);
+    for (bd = c->bd; bd; bd = next) {
+	if (now - bd->last_use > (bd->banmin * 60))
+        {
+		if (bd->next)
+			bd->next->prev = bd->prev;
+		if (bd->prev)
+			bd->prev->next = bd->next;
+		else
+			c->bd = bd->next;
+		if (bd->mask) {
+			send_cmd(s_ChanServ, "%s %s -b %s", 
+			istoken?TOK_MODE:MSG_MODE, c->name, bd->mask);
+			free(bd->mask);
+		}
+		next = bd->next;
+                free(bd);
+		continue;
+	} /* If */
+	next = bd->next;
+    }
+}
+
+void addtban(Channel *c, char *mask, int mins, char *who) {
+
+    BanData *bd, *next;
+    time_t now = time(NULL);
+
+    for (bd = c->bd; bd; bd = next) {
+	if (!stricmp(bd->mask, mask))
+	{
+		bd->last_use = now;
+		return;
+	}
+	next = bd->next;
+    }
+    send_cmd (who, "MODE %s +b %s", c->name, mask);
+    bd = scalloc(sizeof(BanData), 1);
+    bd->mask = sstrdup(mask);
+    bd->last_use = now;
+    bd->banmin = mins;
+
+    bd->prev = NULL;
+    bd->next = c->bd;
+    if (bd->next) bd->next->prev = bd;
+    c->bd = bd;
+    checktban(c);
 }
-#endif
